#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os
import signal
from multiprocessing import Pool

# DIRAC imports
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.setUsageMessage(
    """
Download a list of files or a single file with N parallel process
Usage:
   cta-prod-get-file [options] <lfn> --NbProcess=8
   cta-prod-get-file [options] <ascii file with a list of lfns> --NbProcess=8
"""
)

config = dict(
    n_processes=8,
    skip_existing=False,
)


def set_n_processes(n):
    config["n_processes"] = int(n)


def set_skip(_):
    config["skip_existing"] = True


Script.registerSwitch(
    "n:", "NbProcess=", "number of parallel download processes", set_n_processes
)
Script.registerSwitch("k", "skip-existing", "skip files already downloaded", set_skip)
switches, args = Script.parseCommandLine(ignoreErrors=True)

DIRAC.initialize()

TEMPDIR = ".incomplete"

gLogger.setLevel("NOTICE")


def sigint_handler(signum, frame):
    """
    Raise KeyboardInterrupt on SIGINT (CTRL + C)
    This should be the default, but apparently Dirac changes it.
    """
    raise KeyboardInterrupt()


def getfile(lfn):
    voName = lfn.split("/")[1]
    if voName not in ["ctao", "vo.cta.in2p3.fr"]:
        message = (
            f"Wrong lfn: path must start with vo name (ctao or vo.cta.in2p3.fr):\n{lfn}"
        )
        gLogger.error(message)
        return

    name = os.path.basename(lfn)
    if config["skip_existing"] and os.path.isfile(name):
        gLogger.notice("Skipping already downloaded file", lfn)
        return

    dm = DataManager(vo=voName)
    gLogger.notice("Start downloading file", lfn)
    res = dm.getFile(lfn)

    if not res["OK"]:
        message = res["Message"]
        gLogger.error(f"Error downloading file {lfn}:\n{message}")
        return
    elif res["Value"]["Failed"]:
        message = res["Value"]["Failed"][lfn]
        gLogger.error(f"Error downloading file {lfn}:\n{message}")
        return

    gLogger.info(f"Successfully downloaded file {lfn}")
    os.rename(os.path.join(".incomplete", name), name)


@Script()
def main():
    if len(args) == 1:
        infile = args[0]
    else:
        Script.showHelp()

    # put files currently downloading in a subdirectory
    # to know which files have already finished
    if not os.path.exists(TEMPDIR):
        os.makedirs(TEMPDIR)

    if os.path.isfile(infile):
        infileList = read_inputs_from_file(infile)
    else:
        infileList = [infile]
    # see https://stackoverflow.com/a/35134329/3838691
    # ignore sigint before creating the pool,
    # so child processes inherit the setting, we will terminate them
    # if the master process catches sigint
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    p = Pool(config["n_processes"])
    # add the original handler back
    signal.signal(signal.SIGINT, sigint_handler)

    try:
        future = p.map_async(getfile, infileList)

        while not future.ready():
            future.wait(5)
    except (SystemExit, KeyboardInterrupt):
        gLogger.fatal("Received SIGINT, waiting for subprocesses to terminate")
        p.close()
        p.terminate()
    finally:
        p.close()
        p.join()
        if len(os.listdir(TEMPDIR)) == 0:
            os.rmdir(TEMPDIR)


if __name__ == "__main__":
    main()



from DIRAC.ProductionSystem.DB.ProductionDB import ProductionDB as DIRACProductionDB


class ProductionDB(DIRACProductionDB):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def get_parent_transformation(self, trans_id, connection=False):
        connection = self.__getConnection(connection)
        req = f"SELECT ParentTransformationID FROM ProductionTransformationLinks WHERE TransformationID={trans_id}"
        res = self._query(req, conn=connection)
        return res



from unittest.mock import MagicMock

import pytest
from CTADIRAC.WorkloadManagementSystem.Client.CTAODownloadInputData import (
    CTAODownloadInputData,
)
from DIRAC import S_OK
from DIRAC.WorkloadManagementSystem.Client.test.Test_Client_DownloadInputData import *  # noqa: F403, F405


@pytest.fixture
def mockSE(mocker):
    mockObjectSE = MagicMock()
    mockObjectSE.getFileMetadata.return_value = S_OK(
        {"Successful": {"/a/lfn/1.txt": {"Cached": 1, "Accessible": 1}}, "Failed": {}}
    )
    mockObjectSE.getFile.return_value = S_OK(
        {"Successful": {"/a/lfn/1.txt": {}}, "Failed": {}}
    )
    mockObjectSE.getStatus.return_value = S_OK({"Read": True, "DiskSE": True})
    mockObjectSE.status.return_value = {"Read": True, "DiskSE": True}

    theMockSE = MagicMock()
    theMockSE.return_value = mockObjectSE
    mocker.patch(
        "DIRAC.WorkloadManagementSystem.Client.DownloadInputData.StorageElement",
        new=theMockSE,
    )
    mocker.patch(
        "CTADIRAC.WorkloadManagementSystem.Client.CTAODownloadInputData.StorageElement",
        new=theMockSE,
    )
    return theMockSE


@pytest.fixture
def dli():
    ctao_dli = CTAODownloadInputData(
        {
            "InputData": [],
            "Configuration": {"LocalSEList": ["SE_Local"]},
            "InputDataDirectory": "CWD",
            "FileCatalog": S_OK(
                {
                    "Successful": {
                        "/a/lfn/1.txt": {
                            "Size": 10,
                            "GUID": "aGUID",
                            "SE_Local": "",
                            "SE_Remote": "",
                            "SE_Bad": "",
                            "SE_Tape": "",
                        }
                    }
                }
            ),
        }
    )
    ctao_dli.availableSEs = ["SE_Local", "SE_Remote"]
    return ctao_dli



"""Version information."""
# this is adapted from https://github.com/astropy/astropy/blob/main/astropy/version.py
# see https://github.com/astropy/astropy/pull/10774 for a discussion on why this needed.

try:
    try:
        from ._dev import version
    except Exception:
        from ._version import version
except Exception:
    import warnings

    warnings.warn(
        "Could not determine version; this indicates a broken installation."
        " Install from PyPI, using conda or from a local git repository."
        " Installing github's autogenerated source release tarballs "
        " does not include version information and should be avoided.",
    )
    del warnings
    version = "0.0.0"

__version__ = version
__all__ = ["__version__"]



from pathlib import Path

import pytest
from cwl_utils.parser.cwl_v1_2 import (
    CommandLineTool,
    CommandOutputArraySchema,
    CommandOutputBinding,
    CommandOutputParameter,
    DockerRequirement,
    File,
)

from CTADIRAC.Interfaces.Utilities.CWL_utils import (
    LFN_DIRAC_PREFIX,
    LFN_PREFIX,
    LOCAL_PREFIX,
    extract_and_translate_input_files,
    extract_and_translate_output_files,
    translate_cwl_workflow,
    translate_docker_hints,
    translate_sandboxes_and_lfns,
    verify_cwl_output_type,
)


CVMFS_BASE_PATH = Path("/cvmfs/ctao.dpps.test")


@pytest.fixture
def mock_load_inputfile(mocker):
    return mocker.patch("CTADIRAC.Interfaces.Utilities.CWL_utils.load_inputfile")


@pytest.fixture
def mock_load_document_by_uri(mocker):
    return mocker.patch("CTADIRAC.Interfaces.Utilities.CWL_utils.load_document_by_uri")


@pytest.fixture
def mock_translate_docker_hints(mocker):
    return mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWL_utils.translate_docker_hints"
    )


@pytest.fixture
def mock_extract_and_translate_output_files(mocker):
    return mocker.patch(
        "CTADIRAC.Interfaces.Utilities.CWL_utils.extract_and_translate_output_files"
    )


@pytest.mark.parametrize(
    ("file_input", "expected_result", "expected_sandboxes", "expected_lfns"),
    [
        (
            File(path=LFN_PREFIX + "test_lfn_file.txt"),
            "test_lfn_file.txt",
            [],
            [LFN_DIRAC_PREFIX + "test_lfn_file.txt"],
        ),
        (
            File(path=LOCAL_PREFIX + "test_local_file.txt"),
            "test_local_file.txt",
            ["test_local_file.txt"],
            [],
        ),
        (
            LFN_PREFIX + "test_lfn_str.txt",
            "test_lfn_str.txt",
            [],
            [LFN_DIRAC_PREFIX + "test_lfn_str.txt"],
        ),
        (
            LOCAL_PREFIX + "test_local_str.txt",
            "test_local_str.txt",
            ["test_local_str.txt"],
            [],
        ),
        (File(), None, [], []),  # This will raise an exception
    ],
)
def test_translate_sandboxes_and_lfns(
    file_input, expected_result, expected_sandboxes, expected_lfns
):
    sandboxes = []
    lfns = []

    if expected_result is None:
        with pytest.raises(NotImplementedError, match="File path is not defined."):
            translate_sandboxes_and_lfns(file_input, sandboxes, lfns)
    else:
        result = translate_sandboxes_and_lfns(file_input, sandboxes, lfns)
        assert result == expected_result
        assert sandboxes == expected_sandboxes
        assert lfns == expected_lfns


@pytest.mark.parametrize(
    ("input_data", "expected_result"),
    [
        (
            {"input1": File(path=LFN_PREFIX + "test_lfn_file.txt")},
            {
                "InputDesc": {"input1": File(path="test_lfn_file.txt")},
                "InputSandbox": [],
                "InputData": [LFN_DIRAC_PREFIX + "test_lfn_file.txt"],
            },
        ),
        (
            {"input1": File(path=LOCAL_PREFIX + "test_local_file.txt")},
            {
                "InputDesc": {"input1": File(path="test_local_file.txt")},
                "InputSandbox": ["test_local_file.txt"],
                "InputData": [],
            },
        ),
        (
            {
                "input1": [
                    File(path=LFN_PREFIX + "test_lfn_file1.txt"),
                    File(path=LOCAL_PREFIX + "test_local_file1.txt"),
                ]
            },
            {
                "InputDesc": {
                    "input1": [
                        File(path="test_lfn_file1.txt"),
                        File(path="test_local_file1.txt"),
                    ]
                },
                "InputSandbox": ["test_local_file1.txt"],
                "InputData": [LFN_DIRAC_PREFIX + "test_lfn_file1.txt"],
            },
        ),
        (
            {
                "input1": File(path=LFN_PREFIX + "test_lfn_file2.txt"),
                "input2": File(path=LOCAL_PREFIX + "test_local_file2.txt"),
                "input3": [
                    File(path=LFN_PREFIX + "test_lfn_file3.txt"),
                    File(path=LOCAL_PREFIX + "test_local_file3.txt"),
                ],
            },
            {
                "InputDesc": {
                    "input1": File(path="test_lfn_file2.txt"),
                    "input2": File(path="test_local_file2.txt"),
                    "input3": [
                        File(path="test_lfn_file3.txt"),
                        File(path="test_local_file3.txt"),
                    ],
                },
                "InputSandbox": ["test_local_file2.txt", "test_local_file3.txt"],
                "InputData": [
                    LFN_DIRAC_PREFIX + "test_lfn_file2.txt",
                    LFN_DIRAC_PREFIX + "test_lfn_file3.txt",
                ],
            },
        ),
    ],
)
def test_extract_and_translate_input_files(
    mock_load_inputfile, input_data, expected_result
):
    mock_load_inputfile.return_value = input_data
    result = extract_and_translate_input_files(Path("dummy_path"))
    assert result == expected_result


ARRAY_FILE_OUTPUT = CommandOutputArraySchema(items="test.txt", type_="File")
ARRAY_ARRAY_OUTPUT = CommandOutputArraySchema(
    items=["test.txt", "test2.txt"], type_="array"
)


@pytest.mark.parametrize(
    ("output_type", "expected_result"),
    [
        ("File", True),
        (ARRAY_FILE_OUTPUT, True),
        (ARRAY_ARRAY_OUTPUT, True),
        (["File"], True),
        (["null", "File"], True),
        (["null", ARRAY_FILE_OUTPUT], True),
        (["null", ARRAY_ARRAY_OUTPUT], True),
        ("string", False),
        (["null", "string"], False),
    ],
)
def test_verify_cwl_output_type(output_type, expected_result):
    result = verify_cwl_output_type(output_type)
    assert result is expected_result


@pytest.mark.parametrize(
    ("outputs", "expected_result"),
    [
        (
            [
                CommandOutputParameter(
                    type_="File",
                    outputBinding=CommandOutputBinding(glob="/path/to/output1.txt"),
                )
            ],
            {
                "CWLDesc": {},
                "OutputSandbox": ["/path/to/output1.txt"],
                "OutputData": [],
            },
        ),
        (
            [
                CommandOutputParameter(
                    type_="File",
                    outputBinding=CommandOutputBinding(
                        glob=LFN_PREFIX + "/path/to/output1.txt"
                    ),
                )
            ],
            {
                "CWLDesc": {},
                "OutputSandbox": [],
                "OutputData": [LFN_DIRAC_PREFIX + "/path/to/output1.txt"],
            },
        ),
        (
            [
                CommandOutputParameter(
                    type_=CommandOutputArraySchema(type_="array", items=File),
                    outputBinding=CommandOutputBinding(
                        glob=[
                            LFN_PREFIX + "/path/to/output1.txt",
                            "/path/to/output2.txt",
                        ]
                    ),
                )
            ],
            {
                "CWLDesc": {},
                "OutputSandbox": ["/path/to/output2.txt"],
                "OutputData": [LFN_DIRAC_PREFIX + "/path/to/output1.txt"],
            },
        ),
    ],
)
def test_extract_and_translate_output_files(outputs, expected_result):
    cwl_obj = CommandLineTool(inputs=None, outputs=outputs)
    result = extract_and_translate_output_files(cwl_obj)
    assert result["OutputSandbox"] == expected_result["OutputSandbox"]
    assert result["OutputData"] == expected_result["OutputData"]
    assert isinstance(result["CWLDesc"], CommandLineTool)


@pytest.mark.parametrize(
    ("hints", "base_command", "expected_hints", "expected_base_command"),
    [
        (
            [DockerRequirement(dockerPull="harbor/python:tag")],
            "python",
            [],
            ["apptainer", "run", str(CVMFS_BASE_PATH / "harbor/python:tag"), "python"],
        )
    ],
)
def test_translate_docker_hints(
    hints, base_command, expected_hints, expected_base_command
):
    cwl_obj = CommandLineTool(
        inputs=None, outputs=None, hints=hints, baseCommand="python"
    )
    result = translate_docker_hints(cwl_obj, CVMFS_BASE_PATH, [])
    assert result.hints == expected_hints
    assert result.baseCommand == expected_base_command


@pytest.mark.parametrize(
    ("hints"),
    [
        ([DockerRequirement(dockerPull="harbor/python:tag")],),
    ],
)
def test_translate_cwl_workflow(
    mock_load_document_by_uri,
    mock_translate_docker_hints,
    mock_extract_and_translate_output_files,
    hints,
):
    cwl_obj = CommandLineTool(
        inputs=None, outputs=None, hints=hints, baseCommand="python"
    )
    mock_load_document_by_uri.return_value = cwl_obj
    mock_translate_docker_hints.return_value = cwl_obj
    mock_extract_and_translate_output_files.return_value = {
        "CWLDesc": cwl_obj,
        "OutputSandbox": [],
        "OutputData": [],
    }

    result = translate_cwl_workflow(Path("workflow.cwl"), CVMFS_BASE_PATH, [])

    mock_load_document_by_uri.assert_called_once_with(Path("workflow.cwl"))
    mock_translate_docker_hints.assert_called_once_with(cwl_obj, CVMFS_BASE_PATH, [])
    mock_extract_and_translate_output_files.assert_called_once_with(cwl_obj)
    assert result == {"CWLDesc": cwl_obj, "OutputSandbox": [], "OutputData": []}



#!/usr/bin/env python

"""
  Delete a transformation or a list of transformations

  Usage:
    cta-transformation-delete <list of transID comma separated OR ascii file containing a list of transID (comma-separated on each line)>
"""

__RCSID__ = "$Id$"

import os

import DIRAC
from DIRAC.Core.Base.Script import Script
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


@Script()
def main():
    Script.parseCommandLine()
    args = Script.getPositionalArgs()
    if len(args) != 1:
        Script.showHelp()

    # get arguments
    transIDs = []
    for arg in args[0].split(","):
        if os.path.exists(arg):
            lines = open(arg).readlines()
            for line in lines:
                for transID in line.split(","):
                    transIDs += [int(transID.strip())]
        else:
            transIDs.append(int(arg))

    tc = TransformationClient()

    for transID in transIDs:
        res = tc.deleteTransformation(transID)
        if not res["OK"]:
            DIRAC.gLogger.error(
                f"Failed to delete transformation {transID}: {res['Message']}"
            )
            continue
        else:
            DIRAC.gLogger.notice(f"Successfully deleted transformation {transID}")

    DIRAC.exit(0)


if __name__ == "__main__":
    main()



# Try to use setuptools_scm to get the current version; this is only used
# in development installations from the git repository.
# see ../version.py for details
try:
    from setuptools_scm import get_version

    version = get_version(root="../../..", relative_to=__file__)
except Exception as e:
    raise ImportError(f"setuptools_scm broken or not installed: {e}")



""" Create and send a combined request for any pending operations at
    the end of a job:

      * fileReport (for the transformation)
      * jobReport (for jobs)
      * accounting
      * request (for failover)
"""
from DIRAC import gLogger
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.Workflow.Modules.ModuleBase import ModuleBase, GracefulTermination


class cta_update_ts(ModuleBase):
    #############################################################################

    def __init__(self):
        """Module initialization."""

        self.log = gLogger.getSubLogger(self.__class__.__name__)
        super().__init__(self.log)

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here."""
        super()._resolveInputVariables()
        super()._resolveInputStep()

    def _initialize(self):
        """checks if is to do, then prepare few things"""
        if not self._enableModule():
            raise GracefulTermination("Skipping cta_update_ts module")

        self.request.RequestName = "job_%d_request.xml" % self.jobID
        self.request.JobID = self.jobID
        self.request.SourceComponent = "Job_%d" % self.jobID

    def _execute(self):
        # report on the status of the input data, by default they are 'Processed', unless the job failed
        # failures happening before are not touched
        filesInFileReport = self.fileReport.getFiles()
        if self._checkWFAndStepStatus(noPrint=True):
            filesInFileReport = self.fileReport.getFiles()
            for lfn in self.inputDataList:
                if lfn not in filesInFileReport:
                    self.log.verbose(
                        f"No status populated for input data {lfn}, setting to 'Processed'"
                    )
                    self.fileReport.setFileStatus(
                        int(self.production_id),
                        lfn,
                        TransformationFilesStatus.PROCESSED,
                    )

        result = self.fileReport.commit()
        if not result["OK"]:
            self.log.error("Failed to report file status to TransformationDB")
            self.log.error(
                "Trying again before populating request with file report information"
            )
            result = self.fileReport.generateForwardDISET()
            if not result["OK"]:
                self.log.warn(
                    f"Could not generate Operation for file report with result:\n{result['Value']}"
                )
            else:
                if result["Value"] is None:
                    self.log.info("Files correctly reported to TransformationDB")
                else:
                    result = self.request.addOperation(result["Value"])
        else:
            self.log.info(
                "Status of files have been properly updated in the TransformationDB"
            )

        # Must ensure that the local job report instance is used to report the final status
        # in case of failure and a subsequent failover operation
        if self.workflowStatus["OK"] and self.stepStatus["OK"]:
            self.setApplicationStatus("Job Finished Successfully")

            self.generateFailoverFile()

    def _finalize(self):
        """Finalize and report correct status for the workflow based on the workflow
        or step status.
        """
        if not self._checkWFAndStepStatus(True):
            raise RuntimeError("Workflow failed, cta_update_ts module completed")

        super()._finalize("Workflow successful, end of cta_update_ts module execution.")



#!/usr/bin/env python
"""
Create yml configuration files to submit dl0->dl2 processing workflow with ctapipe

dl0_to_dl2.yml
merge.yml
train_energy.yml
train_classifier.yml
apply_moodels.yml

Then use:
cta-prod-submit <prodName> dl0_to_dl2.yml
cta-prod-submit <prodName> merge.yml
....

Usage example:
   cta-prod-create-workflow-config <workflow_template.yml>

Example of workflow_template:

MCCampaign: PROD6
site: LaPalma
input_array_layout: Prod6-Hyperarray
processing_array_layout: Alpha
zenith: 60.0
analysis_prog_version: v0.23.0
analysis_config_version: v1
configuration_id: 16
moon: dark
pointing: North
"""

__RCSID__ = "$Id$"

from copy import deepcopy
from ruamel.yaml import YAML

import DIRAC
from DIRAC.Core.Base.Script import Script

Script.parseCommandLine()


yaml = YAML()
yaml.indent(mapping=2, sequence=4, offset=2)

arguments = Script.getPositionalArgs()
if len(arguments) != 1:
    Script.showHelp()

workflow_config_file = arguments[0]

with open(workflow_config_file) as stream:
    workflow_config = yaml.load(stream)

mandatory_keys = [
    "MCCampaign",
    "site",
    "input_array_layout",
    "processing_array_layout",
    "zenith",
    "analysis_prog_version",
    "analysis_config_version",
    "moon",
    "pointing",
]

for key in mandatory_keys:
    if key not in workflow_config:
        DIRAC.gLogger.notice(f"Missing mandatory key {key}")
        DIRAC.exit(-1)

MCCampaign = workflow_config["MCCampaign"]
site = workflow_config["site"]

if site not in ["LaPalma", "Paranal"]:
    DIRAC.gLogger.notice(f"Invalid value {site} for site")
    DIRAC.exit(-1)

input_array_layout = workflow_config["input_array_layout"]
processing_array_layout = workflow_config["processing_array_layout"]
moon = workflow_config["moon"]
config_version = workflow_config["analysis_config_version"]

if moon == "dark":
    nsb = "NSB1x"
elif moon == "half":
    nsb = "NSB5x"
elif moon == "full":
    nsb = "NSB19x"
else:
    DIRAC.gLogger.notice(f"Invalid value {moon} for moon")
    DIRAC.exit(-1)
pointing = workflow_config["pointing"]
if pointing == "North":
    phiP = 180.0
elif pointing == "South":
    phiP = 0.0
else:
    DIRAC.gLogger.notice(f"Invalid value {pointing} for pointing")
    DIRAC.exit(-1)

thetaP = workflow_config["zenith"]
zenith_str = str(int(thetaP)) + "deg"

version = workflow_config["analysis_prog_version"]
configuration_id = workflow_config["configuration_id"]

common_dict = {}
common_dict["MCCampaign"] = MCCampaign
common_dict["configuration_id"] = configuration_id
common_dict["base_path"] = "/vo.cta.in2p3.fr/MC"


def create_ctapipe_process_config(dl_level):
    prod_dict = {}
    prod_step_list = []

    id = 0
    for particle in [
        "gamma__",
        "electron__",
        "gamma-diffuse__train_en",
        "gamma-diffuse__train_cl",
        "gamma-diffuse__test",
        "proton__train_cl",
        "proton__test",
    ]:
        split = particle.split("__")[1]
        if split:
            dataset = (
                MCCampaign.capitalize()
                + "_"
                + site
                + "_"
                + input_array_layout
                + "_"
                + nsb
                + "_"
                + particle.split("__")[0]
                + "_"
                + pointing
                + "_"
                + zenith_str
                + "_R1_"
                + split
            )
        else:
            dataset = (
                MCCampaign.capitalize()
                + "_"
                + site
                + "_"
                + input_array_layout
                + "_"
                + nsb
                + "_"
                + particle.split("__")[0]
                + "_"
                + pointing
                + "_"
                + zenith_str
                + "_R1"
            )
        id += 1
        prod_step_dict = {}
        prod_step_dict["ID"] = id
        input_meta_query = {}
        input_meta_query["parentID"] = None
        if dl_level == 1:
            input_meta_query["dataset"] = dataset
        else:
            input_meta_query["parentID"] = None
            input_meta_query["MCCampaign"] = MCCampaign
            input_meta_query["array_layout"] = processing_array_layout
            input_meta_query["site"] = site
            input_meta_query["particle"] = particle.split("_")[0]
            input_meta_query["thetaP"] = thetaP
            input_meta_query["phiP"] = phiP
            input_meta_query["analysis_prog"] = "ctapipe-process"
            input_meta_query["analysis_prog_version"] = version
            input_meta_query["data_level"] = 1
            input_meta_query["outputType"] = "Data"
            input_meta_query["configuration_id"] = configuration_id
            input_meta_query["moon"] = moon
            if split:
                input_meta_query["split"] = split
            else:
                input_meta_query["split"] = "test"
        prod_step_dict["input_meta_query"] = input_meta_query
        job_config = {}
        job_config["type"] = "CtapipeProcessing"
        job_config["version"] = version
        job_config["array_layout"] = processing_array_layout
        job_config["group_size"] = 10
        job_config["output_extension"] = f"DL{dl_level}.h5"
        job_config["data_level"] = dl_level
        job_config["tag"] = "processing"
        if site == "LaPalma":
            subarray_cfg = "subarray_north_" + processing_array_layout.lower() + ".yml"
        elif site == "Paranal":
            subarray_cfg = "subarray_south_" + processing_array_layout.lower() + ".yml"
        if dl_level == 1:
            options = f"-c {config_version}/dl0_to_dl1.yml"
            if not split:
                job_config["split"] = "test"
        else:
            options = f"-c {config_version}/dl1_to_dl2.yml -c {config_version}/{MCCampaign.lower()}/{subarray_cfg}"
        job_config["options"] = options
        prod_step_dict["job_config"] = job_config
        prod_step_list.append(prod_step_dict)

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    if dl_level == 1:
        file_name = "dl0_to_dl1.yml"
    else:
        file_name = "dl1_to_dl2.yml"

    with open(file_name, "w") as file:
        yaml.dump(prod_dict, file)


def create_merge_config():
    prod_dict = {}
    prod_step_list = []

    id = 0
    for particle in [
        "gamma_",
        "electron_",
        "gamma-diffuse_train_en",
        "gamma-diffuse_train_cl",
        "gamma-diffuse_test",
        "proton_train_cl",
        "proton_test",
    ]:
        for i in range(0, 3):
            id += 1
            prod_step_dict = {}
            prod_step_dict["ID"] = id
            input_meta_query = {}

            job_config = {}
            job_config["type"] = "Merging"
            job_config["version"] = version

            if id in [3, 6, 9, 12, 15, 18, 21]:
                job_config["group_size"] = 10000
            else:
                job_config["group_size"] = 25

            if id in [2, 5, 8, 11, 14, 17, 20]:
                input_meta_query["parentID"] = id - 1
                job_config["output_extension"] = "01_merged.DL2.h5"
            elif id in [3, 6, 9, 12, 15, 18, 21]:
                input_meta_query["parentID"] = id - 1
                if particle in [
                    "gamma_",
                    "electron_",
                    "gamma-diffuse_test",
                    "proton_test",
                ]:
                    job_config["output_extension"] = (
                        processing_array_layout.lower() + "_" + "test_merged.DL2.h5"
                    )
                elif particle in [
                    "gamma-diffuse_train_en",
                    "gamma-diffuse_train_cl",
                    "proton_train_cl",
                ]:
                    job_config["output_extension"] = (
                        processing_array_layout.lower()
                        + "_train_"
                        + particle.split("_")[2]
                        + "_merged.DL2.h5"
                    )
            else:
                job_config["output_extension"] = "merged.DL2.h5"
                input_meta_query["parentID"] = None
                input_meta_query["MCCampaign"] = MCCampaign
                input_meta_query["array_layout"] = processing_array_layout
                input_meta_query["site"] = site
                input_meta_query["particle"] = particle.split("_")[0]
                input_meta_query["thetaP"] = thetaP
                input_meta_query["phiP"] = phiP
                input_meta_query["analysis_prog"] = "ctapipe-process"
                input_meta_query["analysis_prog_version"] = version
                input_meta_query["data_level"] = 2
                input_meta_query["outputType"] = "Data"
                input_meta_query["configuration_id"] = configuration_id
                input_meta_query["merged"] = 0
                input_meta_query["moon"] = moon
                for split in ["train_en", "train_cl", "test"]:
                    if split in particle:
                        input_meta_query["split"] = split

            prod_step_dict["input_meta_query"] = input_meta_query
            prod_step_dict["job_config"] = job_config
            prod_step_list.append(deepcopy(prod_step_dict))

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    with open("merge.yml", "w") as file:
        yaml.dump(prod_dict, file)


def create_train_energy_config():
    prod_dict = {}
    prod_step_list = []
    prod_step_dict = {}
    input_meta_query = {}

    prod_step_dict["ID"] = 1
    input_meta_query["parentID"] = None
    input_meta_query["MCCampaign"] = MCCampaign
    input_meta_query["array_layout"] = processing_array_layout
    input_meta_query["site"] = site
    input_meta_query["particle"] = "gamma-diffuse"
    input_meta_query["split"] = "train_en"
    input_meta_query["thetaP"] = thetaP
    input_meta_query["phiP"] = phiP
    input_meta_query["analysis_prog"] = "ctapipe-merge"
    input_meta_query["analysis_prog_version"] = version
    input_meta_query["data_level"] = 2
    input_meta_query["outputType"] = "Data"
    input_meta_query["configuration_id"] = configuration_id
    input_meta_query["merged"] = 3
    input_meta_query["moon"] = moon
    prod_step_dict["input_meta_query"] = input_meta_query
    job_config = {}
    job_config["type"] = "CtapipeTrainEnergy"
    job_config["version"] = version
    job_config["tag"] = "training"
    options = f"-c {config_version}/train_energy_regressor.yml"
    job_config["options"] = options
    prod_step_dict["job_config"] = job_config
    prod_step_list.append(prod_step_dict)

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    with open("train_energy.yml", "w") as file:
        yaml.dump(prod_dict, file)


def create_train_classifier_config():
    prod_dict = {}
    prod_step_list = []

    id = 0
    for particle in ["gamma-diffuse", "proton"]:
        id += 1

        prod_step_dict = {}
        input_meta_query = {}

        prod_step_dict["ID"] = id
        input_meta_query["parentID"] = None
        input_meta_query["MCCampaign"] = MCCampaign
        input_meta_query["array_layout"] = processing_array_layout
        input_meta_query["site"] = site
        input_meta_query["particle"] = particle
        input_meta_query["split"] = "train_cl"
        input_meta_query["thetaP"] = thetaP
        input_meta_query["phiP"] = phiP
        input_meta_query["analysis_prog"] = "ctapipe-merge"
        input_meta_query["analysis_prog_version"] = version
        input_meta_query["data_level"] = 2
        input_meta_query["outputType"] = "Data"
        input_meta_query["configuration_id"] = configuration_id
        input_meta_query["merged"] = 3
        input_meta_query["moon"] = moon
        prod_step_dict["input_meta_query"] = input_meta_query
        job_config = {}
        job_config["type"] = "CtapipeApplyModels"
        job_config["version"] = version
        job_config["options"] = "--reconstructor energy_model.pkl"
        prod_step_dict["job_config"] = job_config
        prod_step_list.append(prod_step_dict)

    id += 1

    prod_step_dict = {}
    input_meta_query = {}

    prod_step_dict["ID"] = id
    input_meta_query["parentID"] = None
    input_meta_query["MCCampaign"] = MCCampaign
    input_meta_query["array_layout"] = processing_array_layout
    input_meta_query["site"] = site
    input_meta_query["split"] = "train_cl"
    input_meta_query["thetaP"] = thetaP
    input_meta_query["phiP"] = phiP
    input_meta_query["analysis_prog"] = "ctapipe-apply-models"
    input_meta_query["analysis_prog_version"] = version
    input_meta_query["data_level"] = 2
    input_meta_query["outputType"] = "Data"
    input_meta_query["configuration_id"] = configuration_id
    input_meta_query["moon"] = moon
    prod_step_dict["input_meta_query"] = input_meta_query
    job_config = {}
    job_config["type"] = "CtapipeTrainClassifier"
    job_config["version"] = version
    options = f"-c {config_version}/train_particle_classifier.yml"
    job_config["options"] = options
    job_config["tag"] = "training"
    job_config["group_size"] = 2
    prod_step_dict["job_config"] = job_config
    prod_step_list.append(prod_step_dict)

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    with open("train_classifier.yml", "w") as file:
        yaml.dump(prod_dict, file)


def create_apply_models_config():
    prod_dict = {}
    prod_step_list = []

    id = 0
    for particle in ["gamma", "electron", "gamma-diffuse", "proton"]:
        id += 1

        prod_step_dict = {}
        input_meta_query = {}

        prod_step_dict["ID"] = id
        input_meta_query["parentID"] = None
        input_meta_query["MCCampaign"] = MCCampaign
        input_meta_query["array_layout"] = processing_array_layout
        input_meta_query["site"] = site
        input_meta_query["particle"] = particle
        input_meta_query["split"] = "test"
        input_meta_query["thetaP"] = thetaP
        input_meta_query["phiP"] = phiP
        input_meta_query["analysis_prog"] = "ctapipe-merge"
        input_meta_query["analysis_prog_version"] = version
        input_meta_query["data_level"] = 2
        input_meta_query["outputType"] = "Data"
        input_meta_query["configuration_id"] = configuration_id
        input_meta_query["merged"] = 3
        input_meta_query["moon"] = moon
        prod_step_dict["input_meta_query"] = input_meta_query
        job_config = {}
        job_config["type"] = "CtapipeApplyModels"
        job_config["version"] = version
        job_config[
            "options"
        ] = "--reconstructor energy_model.pkl --reconstructor classifier_model.pkl"
        prod_step_dict["job_config"] = job_config
        prod_step_list.append(prod_step_dict)

    prod_dict["ProdSteps"] = prod_step_list
    prod_dict["Common"] = common_dict

    with open("apply_models.yml", "w") as file:
        yaml.dump(prod_dict, file)


@Script()
def main():
    create_ctapipe_process_config(1)
    create_ctapipe_process_config(2)
    create_merge_config()
    create_train_energy_config()
    create_train_classifier_config()
    create_apply_models_config()


########################################################
if __name__ == "__main__":
    main()



"""Helper module that contains a set of useful simple and generic functions

JB, September 2018
"""

__RCSID__ = "$Id$"

import copy
import datetime
import os
import re

import DIRAC
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient

# Jobs status dictionnary
BASE_STATUS_DIR = {
    "Received": 0,
    "Matched": 0,
    "Waiting": 0,
    "Running": 0,
    "Failed": 0,
    "Stalled": 0,
    "Rescheduled": 0,
    "Checking": 0,
    "Done": 0,
    "Completing": 0,
    "Completed": 0,
    "Killed": 0,
    "Total": 0,
}

# Data level meta data id
DATA_LEVEL_METADATA_ID = {
    "MC0": -3,
    "R0": -2,
    "R1": -1,
    "DL0": 0,
    "DL1": 1,
    "DL2": 2,
    "DL3": 3,
    "DL4": 4,
    "DL5": 5,
}


def highlight(string):
    """highlight a string in a terminal display"""
    return f"\x1b[31;1m{string}\x1b[0m"


def read_inputs_from_file(file_path):
    """Read inputs from an ASCII files
    Used for list of LFN or list of datasets
    Expects just one LFN/dataset per line
    """
    content = open(file_path).readlines()
    input_file_list = []
    for line in content:
        infile = line.strip()
        if line != "\n":
            input_file_list.append(infile)
    return input_file_list


def extract_run_number_from_logs_tgz(filename):
    match = re.search(r"_(\d+)\.logs\.tgz$", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_chimp_mars(filename):
    match = re.search(r"run(\d+)___cta", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_corsika_sim_telarray(filename):
    if filename.endswith(".corsika.zst"):
        match = re.search(r"run(\d+)_", os.path.basename(filename))
    elif "tid" in filename:
        match = re.search(r"tid(\d+)", os.path.basename(filename))
    elif filename.endswith(".log"):
        match = re.search(r"run(\d+)\.log$", os.path.basename(filename))
    elif filename.endswith(".corsika.log.gz"):
        match = re.search(r"run(\d+)", os.path.basename(filename))
    else:
        match = re.search(r"run(\d+)___cta", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_evndisplay(filename):
    if "tid" in filename:
        match = re.search(r"tid(\d+)", os.path.basename(filename))
    elif filename.endswith(("DL1.root", "DL2.root", "DL1.tar.gz", "DL2.tar.gz")):
        match = re.search(r"run(\d+)___cta", filename)
    else:
        match = re.search(r"(\d+)-", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_image_extractor(filename):
    match = re.search(r"srun(\d+)-", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_dl1_data_handler(filename):
    match = re.search(r"runs(\d+)-", filename)
    return int(match.group(1)) if match else -1


def extract_run_number_from_ctapipe(filename):
    match = re.search(r"run(\d+)___cta", filename)
    return int(match.group(1)) if match else -1


def run_number_from_filename(filename, package):
    """Try to get a run number from the file name

    return:
        run_number : int - the run number
    """
    if filename.endswith(".logs.tgz"):
        return extract_run_number_from_logs_tgz(filename)

    package_extractors = {
        "chimp": extract_run_number_from_chimp_mars,
        "mars": extract_run_number_from_chimp_mars,
        "corsika_simhessarray": extract_run_number_from_corsika_sim_telarray,
        "corsika_simtelarray": extract_run_number_from_corsika_sim_telarray,
        "evndisplay": extract_run_number_from_evndisplay,
        "image_extractor": extract_run_number_from_image_extractor,
        "dl1_data_handler": extract_run_number_from_dl1_data_handler,
        "ctapipe_dl1": extract_run_number_from_ctapipe,
        "ctapipe_dl2": extract_run_number_from_ctapipe,
    }

    if package in package_extractors:
        return package_extractors[package](filename)

    return -1


def check_dataset_query(dataset_name):
    """print dfind command for a given dataset"""
    md_dict = get_dataset_MQ(dataset_name)
    return debug_query(md_dict)


def debug_query(MDdict):
    """just unwrap a meta data dictionnary into a dfind command"""
    msg = "dfind /vo.cta.in2p3.fr/MC/"
    for key, val in MDdict.items():
        try:
            val = val.values()[0]
        except BaseException:
            pass
        msg += f" {key}={val}"
    return msg


def get_dataset_MQ(dataset_name):
    """Return the Meta Query associated with a given dataset"""
    fc = FileCatalogClient()
    result = returnSingleResult(fc.getDatasetParameters(dataset_name))
    if not result["OK"]:
        DIRAC.gLogger.error("Failed to retrieved dataset:", result["Message"])
        DIRAC.exit(-1)
    else:
        DIRAC.gLogger.info("Successfully retrieved dataset: ", dataset_name)
    return result["Value"]["MetaQuery"]


def get_job_list(owner, job_group, n_hours):
    """get a list of jobs for a selection"""
    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()

    now = datetime.datetime.now()
    onehour = datetime.timedelta(hours=1)
    results = dirac.selectJobs(
        jobGroup=job_group, owner=owner, date=now - n_hours * onehour
    )
    if "Value" not in results:
        DIRAC.gLogger.error(
            'No job found for group "%s" and owner "%s" in the past %s hours'
            % (job_group, owner, n_hours)
        )
        DIRAC.exit(-1)

    # Found some jobs, print information)
    jobs_list = results["Value"]
    return jobs_list


def parse_jobs_list(jobs_list):
    """parse a jobs list by first getting the status of all jobs"""
    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    # status of all jobs
    status = dirac.getJobStatus(jobs_list)
    # parse it
    sites_dict = {}
    status_dict = copy.copy(BASE_STATUS_DIR)
    for job in jobs_list:
        site = status["Value"][int(job)]["Site"]
        majstatus = status["Value"][int(job)]["Status"]
        if majstatus not in status_dict.keys():
            DIRAC.gLogger.notice(f"Add {majstatus} to BASE_STATUS_DIR")
            DIRAC.sys.exit(1)
        status_dict[majstatus] += 1
        status_dict["Total"] += 1
        if site not in sites_dict.keys():
            if site.find(".") == -1:
                site = "    None"  # note that blank spaces are needed
            sites_dict[site] = copy.copy(BASE_STATUS_DIR)
            sites_dict[site][majstatus] = 1
            sites_dict[site]["Total"] = 1
        else:
            sites_dict[site]["Total"] += 1
            if majstatus not in sites_dict[site].keys():
                sites_dict[site][majstatus] = 1
            else:
                sites_dict[site][majstatus] += 1
    return status_dict, sites_dict


def get_os_and_cpu_info():
    """get OS and instructions supported by current cpu"""
    import platform
    import re
    import subprocess

    platform = platform.platform()
    cpuinfo = subprocess.check_output("cat /proc/cpuinfo", shell=True, text=True)
    model_name = subprocess.check_output(
        'grep -m 1 "model name" /proc/cpuinfo', shell=True, text=True
    ).strip()
    for inst in ["avx512", "avx2", "avx", "sse4"]:
        if re.search(inst, cpuinfo) is not None:
            break
    return (platform, model_name, inst)



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os
from multiprocessing import Pool

# DIRAC imports
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.setUsageMessage(
    """
Bulk upload of a list of local files or a single file from the current directory to a Storage Element

lfns must contain local file names, i.e.:
lfn = /ctao/<subdirs>/<local_file>

Usage:
   cta-prod-add-file <lfn> <SE>
   cta-prod-add-file <ascii file with a list of lfns> <SE>
"""
)

Script.parseCommandLine(ignoreErrors=True)

args = Script.getPositionalArgs()
if len(args) > 1:
    infile = args[0]
    SE = args[1]
else:
    Script.showHelp()

DIRAC.initialize()

gLogger.setLevel("NOTICE")

res = getProxyInfo()

if not res["OK"]:
    gLogger.error("Error getting proxy info")
    DIRAC.exit(2)

voName = res["Value"]["VOMS"][0].split("/")[1]


@Script()
def main():
    if os.path.isfile(infile):
        infileList = read_inputs_from_file(infile)
    else:
        infileList = [infile]

    p = Pool(10)
    p.map(addfile, infileList)


def addfile(lfn):
    start_path = lfn.split("/")[1]
    if start_path != voName:
        gLogger.error(f"Wrong lfn: path must start with vo name {voName}:\n {lfn}")
        return
    dm = DataManager(vo=voName)
    res = dm.putAndRegister(lfn, os.path.basename(lfn), SE)
    if not res["OK"]:
        message = res["Message"]
        gLogger.error(f"Error uploading file {lfn}:\n{message}")
        return
    gLogger.info(f"Successfully uploaded file {lfn}")


if __name__ == "__main__":
    main()



import xml.etree.ElementTree as ET
from pathlib import Path

import pytest

from CTADIRAC.Interfaces.API.CWLJob import CWLJob
from CTADIRAC.Interfaces.Utilities.CWL_utils import (
    LFN_DIRAC_PREFIX,
    LFN_PREFIX,
)

INPUT_DATA = ["/ctao/user/MC/prod.sim", "/ctao/user/MC/prod2.sim"]
INPUT_SANDBOX = ["/a/local/MC/simulation.py", "/path/to/MC.prod3.sim"]
CWL_INPUTS_EXAMPLE = f"""
local_script:
  - class: File
    path: {INPUT_SANDBOX[0]}
input_as_lfn:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[0]}
input_as_lfn_2:
  - class: File
    path: {LFN_PREFIX}{INPUT_DATA[1]}
  - class: File
    path: {INPUT_SANDBOX[1]}
dataset: "dataset://path/to/data"
input_param: "a random param"
"""
OUTPUT_DATA = ["/ctao/user/MC/fit*.out", "/ctao/usr/MC/data.sim"]
OUTPUT_SANDBOX = ["/path/to/test*.out", "/path/to/data_2.sim", "*.txt"]
BASE_COMMAND = "python"
IMAGE = "harbor.cta-observatory.org/proxy_cache/library/python:3.12-slim"
CWL_WORKFLOW_EXAMPLE = f"""
cwlVersion: v1.2
class: CommandLineTool
doc: |
      Test using local input/output data treated as Dirac
      input/output sandboxes.

inputs:
  local_script:
    type: File
    inputBinding:
      position: 1
  input_as_lfn:
    type: File?
    inputBinding:
      position: 2
  input_as_lfn_2:
    type: File[]
    inputBinding:
      position: 3
  dataset:
    type: [File, string]
    inputBinding:
      position: 4
  input_param:
    type: string
    inputBinding:
      position: 5

outputs:
  output_as_sb:
    type: File[]?
    outputBinding:
      glob: ["{OUTPUT_SANDBOX[0]}"]
  output_as_lfn:
    type: File?
    label: "LFN wildcards"
    outputBinding:
      glob: "{LFN_PREFIX}{OUTPUT_DATA[0]}"
  output_as_lfn_2:
    type: File[]
    label: "LFN files list"
    outputBinding:
      glob:
        - {LFN_PREFIX}{OUTPUT_DATA[1]}
        - {OUTPUT_SANDBOX[1]}
  output_as_array:
    type:
      type: array
      items: File
    outputBinding:
      glob: "{OUTPUT_SANDBOX[2]}"

baseCommand: ["{BASE_COMMAND}"]

hints:
  DockerRequirement:
    dockerPull: {IMAGE}
"""


@pytest.fixture
def mock_submit_job(mocker):
    return mocker.patch(
        "CTADIRAC.Interfaces.API.CWLJob.Dirac.submitJob",
        side_effect=lambda self: self._toXML(),
    )


@pytest.mark.parametrize(
    (
        "cwl_worflow",
        "cwl_inputs",
        "expected_input_data",
        "expected_input_sandbox",
        "expected_output_data",
        "expected_output_sandbox",
    ),
    [
        (
            CWL_WORKFLOW_EXAMPLE,
            CWL_INPUTS_EXAMPLE,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in INPUT_DATA],
            INPUT_SANDBOX,
            [f"{LFN_DIRAC_PREFIX}{data}" for data in OUTPUT_DATA],
            OUTPUT_SANDBOX,
        ),
    ],
)
def test_cwl_job_submit(
    mock_submit_job,
    tmp_path,
    cwl_worflow,
    cwl_inputs,
    expected_input_data,
    expected_input_sandbox,
    expected_output_data,
    expected_output_sandbox,
):
    cwl_workflow_example = tmp_path / "cwl_workflow_example.cwl"
    cwl_workflow_example.write_text(cwl_worflow)
    cwl_inputs_example = tmp_path / "cwl_inputs_example.cwl"
    cwl_inputs_example.write_text(cwl_inputs)

    job = CWLJob(
        cwl_workflow=cwl_workflow_example,
        cwl_inputs=cwl_inputs_example,
        cvmfs_base_path=Path("/cvmfs/ctao.dpps.test/"),
    )

    result = job.submit()
    mock_submit_job.assert_called_once_with(job)
    result_xml = ET.fromstring(result)
    input_data = result_xml.find(".//Parameter[@name='InputData']/value").text.split(
        ";"
    )
    assert input_data == expected_input_data
    output_sandbox = result_xml.find(
        ".//Parameter[@name='OutputSandbox']/value"
    ).text.split(";")
    assert set(expected_output_sandbox).issubset(output_sandbox)
    input_sandbox = result_xml.find(
        ".//Parameter[@name='InputSandbox']/value"
    ).text.split(";")
    assert set(expected_input_sandbox).issubset(input_sandbox)
    output_data = result_xml.find(".//Parameter[@name='OutputData']/value").text.split(
        ";"
    )
    assert output_data == expected_output_data
    executable = result_xml.find(
        ".//StepInstance/Parameter[@name='executable']/value"
    ).text
    assert executable == "cwltool"



"""Utilities to handle CWL."""

from pathlib import Path
from typing import Any

from cwl_utils.parser import load_document_by_uri
from cwl_utils.parser.cwl_v1_2 import CommandOutputArraySchema, File
from cwl_utils.parser.cwl_v1_2_utils import load_inputfile

LFN_PREFIX = "lfn://"
LFN_DIRAC_PREFIX = "LFN:"
LOCAL_PREFIX = "file://"


def translate_cwl_workflow(
    cwl_workflow: Path, cvmfs_base_path: Path, apptainer_options: list[Any]
) -> dict[str, Any]:
    """Translate the CWL workflow description into Dirac compliant execution."""
    cwl_obj = load_document_by_uri(cwl_workflow)
    output_dict = {"CWLDesc": cwl_obj, "OutputSandbox": [], "OutputData": []}
    if cwl_obj.class_ != "CommandLineTool":
        return output_dict

    cwl_obj = translate_docker_hints(cwl_obj, cvmfs_base_path, apptainer_options)

    return extract_and_translate_output_files(cwl_obj)


def translate_docker_hints(
    cwl_obj, cvmfs_base_path: Path, apptainer_options: list[Any]
):
    """Translate CWL DockerRequirement into Dirac compliant execution."""
    for index, hints in enumerate(cwl_obj.hints):
        if hints.class_ == "DockerRequirement":
            image = hints.dockerPull
            cmd = [
                "apptainer",
                "run",
                *apptainer_options,
                str(cvmfs_base_path / f"{image}"),
            ]
            if isinstance(cwl_obj.baseCommand, str):
                cmd += [cwl_obj.baseCommand]
            else:
                cmd += cwl_obj.baseCommand
            cwl_obj.baseCommand = cmd
            del cwl_obj.hints[index]
    return cwl_obj


def extract_and_translate_output_files(cwl_obj) -> dict[str, Any]:
    """Translate output files into a DIRAC compliant usage.

    Extract local outputs and lfns.
    Remove outputs path prefix.
    """
    output_lfns = []
    output_sandboxes = []
    for index, outputs in enumerate(cwl_obj.outputs):
        if not verify_cwl_output_type(outputs.type_):
            continue
        glob = outputs.outputBinding.glob
        if isinstance(glob, str):
            cwl_obj.outputs[index].outputBinding.glob = translate_sandboxes_and_lfns(
                glob, output_sandboxes, output_lfns
            )
        if isinstance(glob, list):
            for sub_index, glob_item in enumerate(glob):
                cwl_obj.outputs[index].outputBinding.glob[
                    sub_index
                ] = translate_sandboxes_and_lfns(
                    glob_item, output_sandboxes, output_lfns
                )
    return {
        "CWLDesc": cwl_obj,
        "OutputSandbox": output_sandboxes,
        "OutputData": output_lfns,
    }


def verify_cwl_output_type(output_type) -> bool:
    """Verify the cwl output type.

    True if the type is File or CommandOutputArraySchema
    or a list of 'null' and File/CommandOutputArraySchema
    else False.
    """
    if isinstance(output_type, list):
        for type in output_type:
            if type == "File" or isinstance(type, CommandOutputArraySchema):
                return True
    if output_type == "File" or isinstance(output_type, CommandOutputArraySchema):
        return True
    return False


# TODO: how to deal with default values not present in the input file but in the cwl?
def extract_and_translate_input_files(cwl_inputs: Path) -> dict[str, Any]:
    """Extract input files from CWL inputs, rewrite file prefix.

    If the file is a Sandbox, ensure there is no absolute path,
    and store it in the input sandbox list.
    If the file is a LFN, remove the lfn prefix and store it in the lfns list.
    """
    input_sandboxes = []
    input_lfns = []
    input_obj = load_inputfile(str(cwl_inputs))
    for _, input_value in input_obj.items():
        if isinstance(input_value, list):
            for file in input_value:
                if isinstance(file, File):
                    file.path = translate_sandboxes_and_lfns(
                        file, input_sandboxes, input_lfns
                    )
        elif isinstance(input_value, File):
            input_value.path = translate_sandboxes_and_lfns(
                input_value, input_sandboxes, input_lfns
            )
    return {
        "InputDesc": input_obj,
        "InputSandbox": input_sandboxes,
        "InputData": input_lfns,
    }


def translate_sandboxes_and_lfns(file: File | str, sandboxes: list, lfns: list) -> str:
    """Extract local files as sandboxes and lfns as input data."""
    if isinstance(file, File):
        if not file.path:
            raise NotImplementedError("File path is not defined.")
        path = file.path
    elif isinstance(file, str):
        path = file
    if LFN_PREFIX in path:
        path = path.replace(LFN_PREFIX, "")
        lfns.append(LFN_DIRAC_PREFIX + path)
        return Path(path).name
    path = path.replace(LOCAL_PREFIX, "")
    sandboxes.append(path)
    return Path(path).name



""" The Script class provides a simple way for users to specify an executable
    or file to run (and is also a simple example of a workflow module).
"""
import os
import re
import shlex
import shutil
import stat
import sys

from DIRAC import gLogger, gConfig
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.WorkloadManagementSystem.Utilities.RemoteRunner import RemoteRunner
from DIRAC.Workflow.Modules.ModuleBase import ModuleBase, GracefulTermination


class cta_script(ModuleBase):
    """Module for running executable"""

    #############################################################################
    def __init__(self, log=None):
        """c'tor"""
        if log is not None:
            self.log = log
        else:
            self.log = gLogger.getSubLogger(self.__class__.__name__)
        super().__init__(self.log)

        # Set defaults for all workflow parameters here
        self.executable = ""
        self.applicationName = ""
        self.applicationVersion = ""
        self.applicationLog = ""
        self.arguments = ""
        self.workflow_commons = None
        self.step_commons = None

        self.environment = None
        self.callbackFunction = None
        self.bufferLimit = 52428800

    #############################################################################

    def _resolveInputVariables(self):
        """By convention the workflow parameters are resolved here."""
        super()._resolveInputVariables()
        super()._resolveInputStep()

        self.arguments = self.step_commons.get("arguments", self.arguments)
        if not self.arguments.strip():
            self.arguments = self.workflow_commons.get("arguments", self.arguments)

    #############################################################################

    def _initialize(self):
        """check if the previous step failed"""
        if not self._checkWFAndStepStatus(noPrint=False):
            if self.workflowStatus["Errno"] == 166:
                self.log.info("Workflow failed with status 166 : input files partially processed")
            elif self.workflowStatus["Errno"] == 167:
                self.log.info("Workflow failed with status 167: no input file processed")
            else:
                raise GracefulTermination("Previous step failed : skipping this module")

        """simple checks"""
        if not self.executable:
            raise RuntimeError("No executable defined")

    def _setCommand(self):
        """set the command that will be executed"""
        self.command = self.executable
        if os.path.exists(os.path.basename(self.executable)):
            self.executable = os.path.basename(self.executable)
            if not os.access(f"{os.getcwd()}/{self.executable}", 5):
                # doc in https://docs.python.org/2/library/stat.html#stat.S_IRWXU
                os.chmod(
                    f"{os.getcwd()}/{self.executable}",
                    stat.S_IRWXU
                    | stat.S_IRGRP
                    | stat.S_IXGRP
                    | stat.S_IROTH
                    | stat.S_IXOTH,
                )
            self.command = f"{os.getcwd()}/{self.executable}"
        elif re.search(".py$", self.executable):
            self.command = f"{sys.executable} {self.executable}"
        elif shutil.which(self.executable):
            self.command = self.executable

        if self.arguments:
            self.command = f"{self.command} {self.arguments}"

        self.log.info(f"Command is: {self.command}")

    def _executeCommand(self):
        """execute the self.command (uses systemCall)"""
        failed = False

        # Check whether the execution should be done remotely
        if gConfig.getValue("/LocalSite/RemoteExecution", False):
            remoteRunner = RemoteRunner(
                gConfig.getValue("/LocalSite/Site"),
                gConfig.getValue("/LocalSite/GridCE"),
                gConfig.getValue("/LocalSite/CEQueue"),
            )
            retVal = remoteRunner.execute(self.command)
        else:
            retVal = systemCall(
                timeout=0,
                cmdSeq=shlex.split(self.command),
                env=self.environment,
                callbackFunction=self.callbackFunction,
                bufferLimit=self.bufferLimit,
            )

        if not retVal["OK"]:
            failed = True
            self.log.error(
                "System call execution failed:", "\n" + str(retVal["Message"])
            )
            self._exitWithError(1000)
        status, stdout, stderr = retVal["Value"][0:3]
        if status:
            failed = True
            self.log.error(
                "Non-zero status while executing",
                f"{self.command} exited with status {status}",
            )
        else:
            self.log.info(f"{self.command} execution completed with status {status}")

        self.log.verbose(stdout)
        self.log.verbose(stderr)
        if os.path.exists(self.applicationLog):
            self.log.verbose(f"Removing existing {self.applicationLog}")
            os.remove(self.applicationLog)
        with open(f"{os.getcwd()}/{self.applicationLog}", "w") as fopen:
            fopen.write(
                f"<<<<<<<<<< {self.executable} Standard Output >>>>>>>>>>\n\n{stdout} "
            )
            if stderr:
                fopen.write(
                    f"<<<<<<<<<< {self.executable} Standard Error >>>>>>>>>>\n\n{stderr} "
                )
        self.log.info(f"Output written to {self.applicationLog}, execution complete.")

        if failed:
            self._exitWithError(status)

    def _exitWithError(self, status):
        """Here because of possible extensions.

        :param str status: the status of the application becomes the status of the workflow,
                           and may be interpreted by JobWrapper (e.g. for rescheduling cases)
        """
        raise RuntimeError(
            f"'{os.path.basename(self.executable).split('_')[0]}' Exited With Status {status}",
            status,
        )

    def _finalize(self):
        """simply finalize"""
        applicationString = os.path.basename(self.executable).split("_")[0]
        if self.applicationName and self.applicationName.lower() != "unknown":
            applicationString += f" ({self.applicationName} {self.applicationVersion})"
        status = f"{applicationString} successful"

        super()._finalize(status)



#!/usr/bin/env python

__RCSID__ = "$Id$"

# generic imports
import os
from multiprocessing import Pool

# DIRAC imports
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Base.Script import Script
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

from CTADIRAC.Core.Utilities.tool_box import read_inputs_from_file

Script.setUsageMessage(
    """
Bulk removal of a list of files or a single file
Usage:
   cta-prod-remove-file [options] <lfn>
   cta-prod-remove-file [options] <ascii file with lfn list>
"""
)

Script.parseCommandLine(ignoreErrors=True)

DIRAC.initialize()

fc = FileCatalog()


def removeFile(lfn):
    res = fc.removeFile(lfn)
    if not res["OK"]:
        message = res["Message"]
        gLogger.error(f"Error removing file {lfn}:\n{message}")
        return


@Script()
def main():
    args = Script.getPositionalArgs()
    if len(args) > 0:
        infile = args[0]
    else:
        Script.showHelp()

    if os.path.isfile(infile):
        infileList = read_inputs_from_file(infile)
    else:
        infileList = [infile]
    p = Pool(10)
    p.map(removeFile, infileList)


if __name__ == "__main__":
    main()



"""CTAO DIRAC extension package."""


from ._version import __version__

__all__ = [
    "__version__",
]


def extension_metadata():
    return {
        "primary_extension": True,
        "priority": 100,
        "setups": {
            "CTA": "dips://dcta-servers03.pic.es:9135/Configuration/Server,"
            "dips://dcta-agents03.pic.es:9135/Configuration/Server,"
            "dips://ctadirac-01.cscs.cta-observatory.org:9135/Configuration/Server,"
            "dips://ctadirac-02.cscs.cta-observatory.org:9135/Configuration/Server,"
            "dips://ctadirac-03.cscs.cta-observatory.org:9135/Configuration/Server,"
            "dips://ctadirac-04.cscs.cta-observatory.org:9135/Configuration/Server,"
            "dips://ctadirac-05.cscs.cta-observatory.org:9135/Configuration/Server,"
            "dips://cta-dirac.zeuthen.desy.de:9135/Configuration/Server",
            "CTADIRAC-alma": "https://majestix-vm1.zeuthen.desy.de:9135/Configuration/Server",
        },
        "default_setup": "CTA",
    }



""" Download Input Data specific to CTAO
    It derived voName from lfn to download both lfns starting with /vo.cta.in2p3.fr and /ctao
"""
import os
import random

from DIRAC import S_ERROR, S_OK
from DIRAC.Core.Utilities.ReturnValues import returnSingleResult
from DIRAC.Resources.Storage.StorageElement import StorageElement
from DIRAC.WorkloadManagementSystem.Client.DownloadInputData import DownloadInputData

COMPONENT_NAME = "DownloadInputData"


def _isCached(lfn, seName):
    se = StorageElement(seName)
    voName = lfn.split("/")[1]
    se.vo = voName
    result = returnSingleResult(se.getFileMetadata(lfn))
    # result = returnSingleResult(StorageElement(seName).getFileMetadata(lfn))
    if not result["OK"]:
        return False
    metadata = result["Value"]
    return metadata.get("Cached", metadata["Accessible"])


class CTAODownloadInputData(DownloadInputData):
    """
    retrieve InputData LFN from localSEs (if available) or from elsewhere.
    """

    #############################################################################
    def __init__(self, argumentsDict):
        """Standard constructor"""
        DownloadInputData.__init__(self, argumentsDict)

    def execute(self, dataToResolve=None):
        """This method is called to download the requested files in the case where
        enough local disk space is available.  A buffer is left in this calculation
        to leave room for any produced files.
        """

        # Define local configuration options present at every site
        localSESet = set(self.configuration["LocalSEList"])

        if dataToResolve:
            self.log.verbose(
                "Data to resolve passed directly to DownloadInputData module"
            )
            self.inputData = dataToResolve  # e.g. list supplied by another module

        self.inputData = sorted(lfn.replace("LFN:", "") for lfn in self.inputData)
        self.log.info("InputData to be downloaded is:\n%s" % "\n".join(self.inputData))

        replicas = self.fileCatalogResult["Value"]["Successful"]

        # Problematic files will be returned and can be handled by another module
        failedReplicas = set()
        # For the case that a file is found on two SEs at the same site
        # disk-based replicas are favoured.
        downloadReplicas = {}

        for lfn, reps in replicas.items():
            if lfn not in self.inputData:
                self.log.verbose("LFN %s is not in requested input data to download")
                failedReplicas.add(lfn)
                continue

            if not ("Size" in reps and "GUID" in reps):
                self.log.error(f"Missing LFN metadata {lfn} {str(reps)}")
                failedReplicas.add(lfn)
                continue

            # Get and remove size and GUID
            size = reps.pop("Size")
            guid = reps.pop("GUID")
            # Remove all other items that are not SEs
            for item in list(reps):  # note the pop below
                if item not in self.availableSEs:
                    reps.pop(item)
            downloadReplicas[lfn] = {"SE": [], "Size": size, "GUID": guid}
            # First get Disk replicas
            localReps = set(reps) & localSESet
            for seName in localReps:
                seStatus = StorageElement(seName).status()
                if seStatus["DiskSE"] and seStatus["Read"]:
                    downloadReplicas[lfn]["SE"].append(seName)
            # If no disk replicas, take tape replicas
            if not downloadReplicas[lfn]["SE"]:
                for seName in localReps:
                    seStatus = StorageElement(seName).status()
                    if (
                        seStatus["TapeSE"]
                        and seStatus["Read"]
                        and _isCached(lfn, seName)
                    ):
                        # Only consider replicas that are cached
                        downloadReplicas[lfn]["SE"].append(seName)

        totalSize = 0
        verbose = self.log.verbose("Replicas to download are:")
        for lfn, reps in downloadReplicas.items():
            self.log.verbose(lfn)
            if not reps["SE"]:
                self.log.info(
                    f"Failed to find data at local SEs, will try to download from anywhere {lfn}"
                )
                reps["SE"] = ""
            else:
                if len(reps["SE"]) > 1:
                    # if more than one SE is available randomly select one
                    random.shuffle(reps["SE"])
                reps["SE"] = reps["SE"][0]
            totalSize += int(reps.get("Size", 0))
            if verbose:
                for item, value in sorted(reps.items()):
                    if value:
                        self.log.verbose(f"\t{item} {value}")

        self.log.info(f"Total size of files to be downloaded is {totalSize} bytes")
        for lfn in failedReplicas:
            self.log.warn(
                f"Not all file metadata (SE,PFN,Size,GUID) was available for LFN {lfn}"
            )

        # Now need to check that the list of replicas to download fits into
        # the available disk space. Initially this is a simple check and if there is not
        # space for all input data, no downloads are attempted.
        result = self._DownloadInputData__checkDiskSpace(totalSize)
        if not result["OK"]:
            self.log.warn(f"Problem checking available disk space:\n{result}")
            return result

        # FIXME: this can never happen at the moment
        if not result["Value"]:
            self.log.warn(
                f"Not enough disk space available for download {result['Value']} / {totalSize} bytes"
            )
            self._DownloadInputData__setJobParam(
                COMPONENT_NAME,
                f"Not enough disk space available for download: {result['Value']} / {totalSize} bytes",
            )
            return S_OK({"Failed": self.inputData, "Successful": {}})

        resolvedData = {}
        localSECount = 0
        for lfn, info in downloadReplicas.items():
            seName = info["SE"]
            guid = info["GUID"]
            reps = replicas.get(lfn, {})
            if seName:
                se = StorageElement(seName)
                voName = lfn.split("/")[1]
                se.vo = voName
                result = returnSingleResult(se.getFileMetadata(lfn))
                # result = returnSingleResult(StorageElement(seName).getFileMetadata(lfn))
                if not result["OK"]:
                    self.log.error(f"Error getting metadata {result['Message']}")
                    error = result["Message"]
                else:
                    metadata = result["Value"]
                    if metadata.get("Lost", False):
                        error = "PFN has been Lost by the StorageElement"
                    elif metadata.get("Unavailable", False):
                        error = "PFN is declared Unavailable by the StorageElement"
                    elif not metadata.get("Cached", metadata["Accessible"]):
                        error = "PFN is no longer in StorageElement Cache"
                    else:
                        error = ""
                if error:
                    self.log.error(error, lfn)
                    result = {"OK": False}
                else:
                    self.log.info(
                        f"Preliminary checks OK: now downloading {lfn} from {seName}",
                    )
                    result = self._downloadFromSE(lfn, seName, reps, guid)
                    if not result["OK"]:
                        self.log.error(
                            f"Download failed \n Tried downloading from SE {seName}: {result['Message']}"
                        )
                    else:
                        self.log.info(f"Download of {lfn} from {seName} finalized")
            else:
                result = {"OK": False}

            if not result["OK"]:
                reps.pop(seName, None)
                # Check the other SEs
                if reps:
                    self.log.info("Trying to download from any SE")
                    result = self._downloadFromBestSE(lfn, reps, guid)
                    if not result["OK"]:
                        self.log.error(
                            f"Download from best SE failed \n Tried downloading {lfn}: {result['Message']}"
                        )
                        failedReplicas.add(lfn)
                else:
                    failedReplicas.add(lfn)
            else:
                localSECount += 1
            if result["OK"]:
                # Rename file if downloaded FileName does not match the LFN... How can this happen?
                lfnName = os.path.basename(lfn)
                oldPath = result["Value"]["path"]
                fileName = os.path.basename(oldPath)
                if lfnName != fileName:
                    newPath = os.path.join(os.path.dirname(oldPath), lfnName)
                    os.rename(oldPath, newPath)
                    result["Value"]["path"] = newPath
                resolvedData[lfn] = result["Value"]

        # Report datasets that could not be downloaded
        report = ""
        if resolvedData:
            report += f"Successfully downloaded {len(resolvedData)} LFN(s)"
            if localSECount != len(resolvedData):
                report += " (%d from local SEs):\n" % localSECount
            else:
                report += " from local SEs:\n"
            report += "\n".join(sorted(resolvedData))
        failedReplicas = sorted(failedReplicas.difference(resolvedData))
        if failedReplicas:
            self.log.warn(
                f"The following LFN(s) could not be downloaded to the WN:\n{'n'.join(failedReplicas)}"
            )
            report += f"\nFailed to download {len(failedReplicas)} LFN(s):\n"
            report += "\n".join(failedReplicas)

        if report:
            self._DownloadInputData__setJobParam(COMPONENT_NAME, report)

        return S_OK({"Successful": resolvedData, "Failed": failedReplicas})

    def _downloadFromSE(self, lfn, seName, reps, guid):
        """Download a local copy from the specified Storage Element."""
        if not lfn:
            return S_ERROR("LFN not specified: assume file is not at this site")

        self.log.verbose("Attempting to download file", f"{lfn} from {seName}:")

        downloadDir = self._DownloadInputData__getDownloadDir()
        fileName = os.path.basename(lfn)
        for localFile in (
            os.path.join(os.getcwd(), fileName),
            os.path.join(downloadDir, fileName),
        ):
            if os.path.exists(localFile):
                self.log.info(f"File already exists locally {fileName} as {localFile}")
                fileDict = {
                    "turl": "LocalData",
                    "protocol": "LocalData",
                    "se": seName,
                    "pfn": reps[seName],
                    "guid": guid,
                    "path": localFile,
                }
                return S_OK(fileDict)

        localFile = os.path.join(downloadDir, fileName)
        se = StorageElement(seName)
        voName = lfn.split("/")[1]
        se.vo = voName
        result = returnSingleResult(se.getFile(lfn, localPath=downloadDir))
        if not result["OK"]:
            self.log.warn(
                f"Problem getting lfn {lfn} from {seName}:\n{result['Message']}"
            )
            self._DownloadInputData__cleanFailedFile(lfn, downloadDir)
            return result

        if os.path.exists(localFile):
            self.log.verbose(
                "File successfully downloaded locally", f"({lfn} to {localFile})"
            )
            fileDict = {
                "turl": "Downloaded",
                "protocol": "Downloaded",
                "se": seName,
                "pfn": reps[seName],
                "guid": guid,
                "path": localFile,
            }
            return S_OK(fileDict)
        else:
            self.log.warn("File does not exist in local directory after download")
            return S_ERROR("OK download result but file missing in current directory")


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" Manage data and meta-data
"""

__RCSID__ = "$Id$"

# generic imports
import collections
import glob
import json
import os

# DIRAC imports
import DIRAC
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.Utilities import List
from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.TransformationSystem.Client import TransformationFilesStatus
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient


class ProdDataManager:
    """Manage data and meta-data"""

    def __init__(self, catalogs=["DIRACFileCatalog"]):
        """Constructor"""
        self.catalogs = catalogs
        self.setupCatalogClient(catalogs)
        self.printCatalogConfig(catalogs)
        self.transClient = TransformationClient()
        # ## Get the TransformationID
        self.TransformationID = "0000"  # used for local execution or submission to wms
        self.InputData = []
        if "JOBID" in os.environ:
            jobID = os.environ["JOBID"]
            dirac = Dirac()
            res = dirac.getJobJDL(jobID)
            if "TransformationID" in res["Value"]:
                self.TransformationID = res["Value"]["TransformationID"]
            if "InputData" in res["Value"]:
                self.InputData = res["Value"]["InputData"]

    def setupCatalogClient(self, catalogs):
        """Init FileCatalog client
        Ideally we would like to use only FileCatalog but it doesn't work for setMetadata
        because the returned value has a wrong format. So use of FileCatalogClient instead
        """
        self.fc = FileCatalog(catalogs)
        self.fcc = FileCatalogClient()

    def printCatalogConfig(self, catalogs):
        """Dumps catalog configuration"""
        for catalog in catalogs:
            res = self.fc._getCatalogConfigDetails(catalog)
            DIRAC.gLogger.debug("CatalogConfigDetails:", res)

    def _getSEList(self, SEType="ProductionOutputs", DataType="SimtelProd"):
        """get from CS the list of available SE for data upload"""
        opsHelper = Operations()
        optionName = os.path.join(SEType, DataType)
        SEList = opsHelper.getValue(optionName, [])
        SEList = List.randomize(SEList)
        DIRAC.gLogger.notice(f"List of {SEType} SE: {SEList} ")

        # # Check if the local SE is in the list. If yes try it first by reversing list order
        localSEList = []
        res = getSEsForSite(DIRAC.siteName())
        if res["OK"]:
            localSEList = res["Value"]

        retainedlocalSEList = []
        for localSE in localSEList:
            if localSE in SEList:
                DIRAC.gLogger.notice(
                    "The local Storage Element is an available SE: ", localSE
                )
                retainedlocalSEList.append(localSE)
                SEList.remove(localSE)

        SEList = retainedlocalSEList + SEList
        if len(SEList) == 0:
            return DIRAC.S_ERROR("Error in building SEList")

        return DIRAC.S_OK(SEList)

    def _checkemptydir(self, path):
        """check that the directory is not empty"""
        if len(glob.glob(path)) == 0:
            error = f"Empty directory: {path}"
            return DIRAC.S_ERROR(error)
        else:
            return DIRAC.S_OK()

    def _getRunPath(self, filemetadata):
        """format path to string with 1 digit precision
        run_number is taken from filemetadata
        filemetadata can be a dict or the run_number itself
        """
        if isinstance(filemetadata, dict):
            run_number = int(filemetadata["runNumber"])
        else:
            run_number = int(filemetadata)
        run_numberMod = run_number % 1000
        runpath = "%03dxxx" % ((run_number - run_numberMod) / 1000)
        return runpath

    def _formatPath(self, path):
        """format path to string with 1 digit precision"""
        if isinstance(path, float) or isinstance(path, int):
            path = f"{path:.1f}"
        return str(path)

    def createMDTopDirectory(self, metadata, basepath, program_category):
        """create meta data structure"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)

        path = basepath
        process_program = program_category + "_prog"
        for key, value in collections.OrderedDict(
            (k, md[k])
            for k in ("MCCampaign", "site", "particle", process_program)
            if k in md
        ).items():
            path = os.path.join(path, self._formatPath(value))
            res = self.fc.createDirectory(path)
            if not res["OK"]:
                return res

            # Set directory metadata for each subdir: 'site', 'particle', 'process_program'
            res = self.fcc.setMetadata(path, {key: value})
            if not res["OK"]:
                return res

        return DIRAC.S_OK(path)

    def createMDSubDirectory(
        self, metadata, base_path, program_category, output_type_list, trans_id
    ):
        """create meta data structure"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)
        process_program = program_category + "_prog"

        # Create the TransformationID subdir and set MD
        path = os.path.join(base_path, str(trans_id))
        res = self.fc.createDirectory(path)
        if not res["OK"]:
            return res

        process_program_version = process_program + "_version"
        res = self.fcc.setMetadata(
            path,
            {
                k: md[k]
                for k in (
                    "phiP",
                    "thetaP",
                    "array_layout",
                    process_program_version,
                    "sct",
                )
                if k in md
            },
        )
        if not res["OK"]:
            return res

        # Create subdir according to the output_type (Data, Log, Model) and set MD
        transformation_path = path
        for output_type in output_type_list:
            path = os.path.join(transformation_path, output_type)
            res = self.fc.createDirectory(path)
            if not res["OK"]:
                return res

            res = self.fcc.setMetadata(path, {"outputType": output_type})
            if not res["OK"]:
                return res

            # Set additional metadata at subdir level if not already defined
            for metadataKey in [
                "data_level",
                "configuration_id",
            ]:
                res = self.fcc.setMetadata(path, {metadataKey: md[metadataKey]})
                if not res["OK"]:
                    return res

            if output_type in ["Data", "Log"]:
                res = self.fcc.setMetadata(path, {"merged": md.get("merged", 0)})
                if not res["OK"]:
                    return res

        return DIRAC.S_OK(transformation_path)

    def getTransformationPath(self, metadata, basepath, program_category):
        """get the transformation path"""

        ### Create the directory structure
        md = json.loads(metadata, object_pairs_hook=collections.OrderedDict)

        path = basepath
        process_program = program_category + "_prog"
        for value in collections.OrderedDict(
            (k, md[k])
            for k in ("MCCampaign", "site", "particle", process_program)
            if k in md
        ).values():
            path = os.path.join(path, self._formatPath(value))

        # get the TransformationID subdir
        transformation_path = os.path.join(path, self.TransformationID)

        return DIRAC.S_OK(transformation_path)

    def putAndRegister(self, lfn, localfile, filemetadata, DataType):
        """put and register one file and set metadata"""
        # ## Get the list of Production SE
        res = self._getSEList("ProductionOutputs", DataType)
        if res["OK"]:
            ProductionSEList = res["Value"]
        else:
            return res

        # ## Get the list of Failover SE
        res = self._getSEList("ProductionOutputsFailover", DataType)
        if res["OK"]:
            FailoverSEList = res["Value"]
        else:
            return res

        # ## Upload file to a Production SE
        res = self._putAndRegisterToSEList(lfn, localfile, ProductionSEList)
        if not res["OK"]:
            DIRAC.gLogger.error(
                f"Failed to upload file to any Production SE: {ProductionSEList}"
            )
            # ## Upload file to a Failover SE
            res = self._putAndRegisterToSEList(lfn, localfile, FailoverSEList)
            if not res["OK"]:
                return DIRAC.S_ERROR(
                    f"Failed to upload file to any Failover SE: {FailoverSEList}"
                )

        ### Set file metadata also adding jobID metadata
        if res["OK"]:
            fmd = json.loads(filemetadata)
            if "JOBID" in os.environ:
                fmd.update({"jobID": os.environ["JOBID"]})
            res = self.fc.setMetadata(lfn, fmd)
            if not res["OK"]:
                return res

            return DIRAC.S_OK()

    def _putAndRegisterToSEList(self, lfn, localfile, SEList):
        """put and register one file to one SE in the SEList"""
        # Try to upload file to a SE in the list
        voName = lfn.split("/")[1]
        if voName not in ["ctao", "vo.cta.in2p3.fr"]:
            DIRAC.gLogger.error(f"Invalid lfn: path must start with vo name\n {lfn}")
            return DIRAC.S_ERROR()
        dm = DataManager(self.catalogs, vo=voName)

        for SE in SEList:
            msg = f"Try to upload local file: {localfile} \nwith LFN: {lfn} \nto {SE}"
            DIRAC.gLogger.notice(msg)
            res = dm.putAndRegister(lfn, localfile, SE)
            # Check if failed
            if not res["OK"]:
                DIRAC.gLogger.error(
                    f"Failed to putAndRegister {lfn}:\n {res['Message']}"
                )
            else:
                DIRAC.gLogger.notice(f"Successfully uploaded {lfn} to {SE}")
                return DIRAC.S_OK()
        return DIRAC.S_ERROR()

    def setTransformationFileStatus(self, lfn, status):
        if status == "PROCESSED":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.PROCESSED, lfn
            )
        elif status == "PROBLEMATIC":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.PROBLEMATIC, lfn
            )
        elif status == "UNUSED":
            res = self.transClient.setFileStatusForTransformation(
                self.TransformationID, TransformationFilesStatus.UNUSED, lfn
            )
        else:
            DIRAC.gLogger.error(f"Not allowed File Status {status}")
            return DIRAC.S_ERROR()

        return res

    def cleanLocalFiles(self, datadir, pattern="*"):
        """remove files matching pattern in datadir"""

        for localfile in glob.glob(os.path.join(datadir, pattern)):
            DIRAC.gLogger.notice("Removing local file: ", localfile)
            os.remove(localfile)

        return DIRAC.S_OK()



"""DIRAC Job API to execute CWL with cwltool."""

import tempfile
from pathlib import Path

from cwl_utils.parser import save
from DIRAC.Interfaces.API.Dirac import Dirac
from DIRAC.Interfaces.API.Job import Job
from ruamel.yaml import YAML

from CTADIRAC.Interfaces.Utilities.CWL_utils import (
    extract_and_translate_input_files,
    translate_cwl_workflow,
)


class CWLJob(Job):
    """Job class for CWL jobs.
    Submits CommandLineTool using cwltool executable.

    Attrs:
        cwl_workflow: a Path to the local CWL workflow file
        cwl_inputs: a Path to the local CWL inputs file
        cvmfs_base_path: the CVMFS base Path
        apptainer_options: additional apptainer options (default: [])
    """

    def __init__(
        self,
        cwl_workflow: Path,
        cwl_inputs: Path,
        cvmfs_base_path: Path,
        apptainer_options: list | None = None,
    ) -> None:
        super().__init__()
        self.cwl_workflow = cwl_workflow
        self.cwl_inputs = cwl_inputs
        self.cvmfs_base_path = cvmfs_base_path
        self.apptainer_options = apptainer_options if apptainer_options else []

        self.cwl_dict = translate_cwl_workflow(
            self.cwl_workflow, self.cvmfs_base_path, self.apptainer_options
        )
        self.output_sandbox = self.cwl_dict.get("OutputSandbox", [])
        self.output_data = self.cwl_dict.get("OutputData", [])

        self.input_cwl_dict = extract_and_translate_input_files(self.cwl_inputs)
        self.input_data = self.input_cwl_dict.get("InputData", [])
        self.input_sandbox = self.input_cwl_dict.get("InputSandbox", [])

    def submit(self):
        """Submit the CWL job to DIRAC.

        Treat local input and output files as sandbox,
        files starting with 'lfn://' are treated as input/output data
        and translate docker requirements.
        """
        dirac = Dirac()
        yaml = YAML()

        # Create the modified Dirac compliant CWL workflow and inputs files to submit
        with tempfile.NamedTemporaryFile(
            suffix=f"_{self.cwl_workflow.name}"
        ) as temp_workflow:
            yaml.dump(save(self.cwl_dict["CWLDesc"]), temp_workflow)
            temp_workflow.flush()

            with tempfile.NamedTemporaryFile(
                suffix=f"_{self.cwl_inputs.name}"
            ) as temp_inputs:
                yaml.dump(save(self.input_cwl_dict["InputDesc"]), temp_inputs)
                temp_inputs.flush()

                self.setInputSandbox(
                    [str(temp_workflow.file.name), str(temp_inputs.name)]
                    + self.input_sandbox
                )
                self.setOutputSandbox(self.output_sandbox)
                self.setInputData(self.input_data)
                self.setOutputData(self.output_data)

                arguments_str = (
                    f"{Path(temp_workflow.name).name} {Path(temp_inputs.name).name}"
                )

                self.setExecutable(
                    "cwltool", arguments=arguments_str, logFile=f"{self.name}.log"
                )
                res = dirac.submitJob(self)
        return res
